import graphene
from graphene_django.forms.mutation import DjangoModelFormMutation
from graphql_jwt.decorators import login_required, permission_required

from django.utils import timezone
from django.db import transaction

from allianceauth.services.hooks import get_extension_logger

from allianceauth_pve.models import Rotation, Entry, EntryRole, EntryCharacter
from allianceauth_pve.forms import NewRotationForm

from .inputs import EntryInput, RotationCloseInput
from .types import RotationType, EntryType

logger = get_extension_logger(__name__)


class CreateRattingEntry(graphene.Mutation):
    class Arguments:
        input = EntryInput(required=True)
        rotation_id = graphene.Int(required=True)

    ok = graphene.Boolean()
    entry = graphene.Field(EntryType)
    errors = graphene.List(graphene.String)

    @classmethod
    @login_required
    @permission_required('allianceauth_pve.manage_entries')
    def mutate(cls, root, info, input, rotation_id):
        try:
            rotation = Rotation.objects.get(pk=rotation_id)
        except Rotation.DoesNotExist:
            ok = False
            errors = ["Rotation doesn't exists"]
        else:
            ok = not rotation.is_closed
            errors = [] if ok else ['The rotation is closed, you cannot add an entry']

        errors += EntryInput.is_valid(input)
        ok = ok and len(errors) == 0

        if ok:
            with transaction.atomic():
                entry = Entry.objects.create(
                    rotation_id=rotation_id,
                    estimated_total=input['estimated_total'],
                    created_by=info.context.user
                )

                to_add = []

                for new_role in input.roles:
                    to_add.append(EntryRole(
                        entry=entry,
                        name=new_role['name'],
                        value=new_role['value']
                    ))

                EntryRole.objects.bulk_create(to_add)
                to_add.clear()

                setups = set()

                for new_share in input.shares:
                    role = entry.roles.get(name=new_share['role'])

                    setup = new_share['helped_setup'] and new_share['user_id'] not in setups
                    if setup:
                        setups.add(new_share['user_id'])

                    to_add.append(
                        EntryCharacter(
                            entry=entry,
                            role=role,
                            user_character_id=new_share['character_id'],
                            user_id=new_share['user_id'],
                            site_count=new_share['site_count'],
                            helped_setup=setup,
                        )
                    )

                EntryCharacter.objects.bulk_create(to_add)

        else:
            entry = None

        return cls(ok=ok, errors=errors, entry=entry)


class ModifyRattingEntry(graphene.Mutation):
    ok = graphene.Boolean()
    entry = graphene.Field(EntryType)
    errors = graphene.List(graphene.String)

    class Arguments:
        input = EntryInput(required=True)
        entry_id = graphene.Int(required=True)

    @classmethod
    @login_required
    @permission_required('allianceauth_pve.manage_entries')
    def mutate(cls, root, info, input, entry_id):
        errors = EntryInput.is_valid(input)
        ok = len(errors) == 0
        user = info.context.user
        entry = Entry.objects.get(pk=entry_id)

        if entry.rotation.is_closed or (user != entry.created_by and not user.is_superuser):
            ok = False
            errors.append('You cannot edit this entry')

        if ok:
            with transaction.atomic():
                entry.ratting_shares.all().delete()
                entry.roles.all().delete()
                entry.estimated_total = input['estimated_total']
                entry.save()

                to_add = []

                for new_role in input.roles:
                    to_add.append(EntryRole(
                        entry=entry,
                        name=new_role['name'],
                        value=new_role['value']
                    ))

                EntryRole.objects.bulk_create(to_add)
                to_add.clear()

                setups = set()

                for new_share in input.shares:
                    role = entry.roles.get(name=new_share['role'])

                    setup = new_share['helped_setup'] and new_share['user_id'] not in setups
                    if setup:
                        setups.add(new_share['user_id'])

                    to_add.append(EntryCharacter(
                        entry=entry,
                        role=role,
                        user_character_id=new_share['character_id'],
                        user_id=new_share['user_id'],
                        site_count=new_share['site_count'],
                        helped_setup=setup,
                    ))

                EntryCharacter.objects.bulk_create(to_add)

        return cls(ok=ok, errors=errors, entry=entry)


class DeleteRattingEntry(graphene.Mutation):
    ok = graphene.Boolean()
    rotation = graphene.Field(RotationType)

    class Arguments:
        entry_id = graphene.Int(required=True)

    @classmethod
    @login_required
    @permission_required('allianceauth_pve.manage_entries')
    def mutate(cls, root, info, entry_id):
        user = info.context.user
        entry = Entry.objects.select_related('rotation').get(pk=entry_id)
        rotation = entry.rotation

        if rotation.is_closed or (user != entry.created_by and not user.is_superuser):
            ok = False
        else:
            ok = True
            entry.delete()

        return cls(ok=ok, rotation=rotation)


class CreateRotation(DjangoModelFormMutation):
    rotation = graphene.Field(RotationType)

    class Meta:
        form_class = NewRotationForm

    @classmethod
    @login_required
    @permission_required('allianceauth_pve.manage_rotations')
    def perform_mutate(cls, form, info):
        return super().perform_mutate(form, info)


class CloseRotation(graphene.Mutation):
    ok = graphene.Boolean()

    class Arguments:
        input = RotationCloseInput(required=True)

    @classmethod
    @login_required
    @permission_required('allianceauth_pve.manage_rotations')
    def mutate(cls, root, info, input):
        rotation: Rotation = Rotation.objects.get(pk=input.rotation_id)

        if rotation.is_closed:
            ok = False
        else:
            with transaction.atomic():
                rotation.actual_total = input.sales_value
                rotation.is_closed = True
                rotation.closed_at = timezone.now()
                rotation.save()

            ok = True

        return cls(ok=ok)


class Mutation:
    pve_create_entry = CreateRattingEntry.Field()
    pve_modify_entry = ModifyRattingEntry.Field()
    pve_delete_entry = DeleteRattingEntry.Field()
    pve_create_rotation = CreateRotation.Field()
    pve_close_rotation = CloseRotation.Field()



from graphene_django.utils.testing import GraphQLTestCase

from app_utils.testdata_factories import UserMainFactory


class TestTypes(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_zkill_link(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                me {
                    profile {
                        mainCharacter {
                            zkillboard
                        }
                    }
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'me': {
                        'profile': {
                            'mainCharacter': {
                                'zkillboard': f"https://zkillboard.com/character/{self.user.profile.main_character.character_id}/"
                            }
                        }
                    }
                }
            }
        )



import graphene
from graphql_jwt.decorators import login_required, user_passes_test, permission_required

from allianceauth.hrapplications.views import create_application_test
from allianceauth.hrapplications.models import ApplicationForm, Application, ApplicationResponse, ApplicationComment
from allianceauth.notifications import notify

from .inputs import FormAnswerInputType
from .types import ApplicationAdminType, ApplicationType


class CreateApplicationMutation(graphene.Mutation):
    class Arguments:
        input = FormAnswerInputType(required=True)

    ok = graphene.Boolean()
    application = graphene.Field(ApplicationType)

    @classmethod
    @login_required
    @user_passes_test(create_application_test)
    def mutate(cls, root, info, input):
        user = info.context.user
        app_form = ApplicationForm.objects.get(pk=input.form_id)
        if Application.objects.filter(user=user).filter(form=app_form).exists():
            ok = False
            application = None
        else:
            responses = {ans.question_id: ans.answer for ans in input.responses}

            application = Application.objects.create(user=user, form=app_form)
            for question in app_form.questions.all():
                response = ApplicationResponse(question=question, application=application)
                response.answer = "\n".join(responses.get(question.pk, []))
                response.save()

            ok = True

        return cls(ok=ok, application=application)


class DeleteApplicationMutation(graphene.Mutation):
    class Arguments:
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    def mutate(cls, root, info, application_id):
        app = Application.objects.get(pk=application_id)

        if app.user == info.context.user and app.approved is None:
            app.delete()
            ok = True
        else:
            ok = False

        return cls(ok=ok)


class AddApplicationCommentMutation(graphene.Mutation):
    class Arguments:
        comment = graphene.String(required=True)
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()
    application = graphene.Field(ApplicationAdminType)

    @classmethod
    @login_required
    @permission_required('auth.human_resources')
    @permission_required('hrapplications.add_applicationcomment')
    def mutate(cls, root, info, comment, application_id):
        user = info.context.user
        app = Application.objects.get(pk=application_id)

        ApplicationComment.objects.create(
            application=app,
            user=user,
            text=comment,
        )

        return cls(ok=True, application=app)


class AdminRemoveApplicationMutation(graphene.Mutation):
    class Arguments:
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @permission_required('auth.human_resources')
    @permission_required('hrapplications.delete_application')
    def mutate(cls, root, info, application_id):
        app = Application.objects.get(pk=application_id)
        app.delete()
        notify(app.user, "Application Deleted", message=f"Your application to {app.form.corp} was deleted.")
        return cls(ok=True)


class AdminApproveApplication(graphene.Mutation):
    class Arguments:
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()
    application = graphene.Field(ApplicationAdminType)

    @classmethod
    @login_required
    @permission_required('auth.human_resources')
    @permission_required('hrapplications.approve_application')
    def mutate(cls, root, info, application_id):
        user = info.context.user
        app = Application.objects.get(pk=application_id)
        if user.is_superuser or user == app.reviewer:
            app.approved = True
            app.save()
            notify(app.user, "Application Accepted", message="Your application to %s has been approved." % app.form.corp, level="success")
            ok = True
        else:
            ok = False

        return cls(ok=ok, application=app)


class AdminRejectApplication(graphene.Mutation):
    class Arguments:
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()
    application = graphene.Field(ApplicationAdminType)

    @classmethod
    @login_required
    @permission_required('auth.human_resources')
    @permission_required('hrapplications.reject_application')
    def mutate(cls, root, info, application_id):
        user = info.context.user
        app = Application.objects.get(pk=application_id)
        if user.is_superuser or user == app.reviewer:
            app.approved = False
            app.save()
            notify(app.user, "Application Accepted", message="Your application to %s has been approved." % app.form.corp, level="success")
            ok = True
        else:
            ok = False

        return cls(ok=ok, application=app)


class AdminMarkInProgressApplication(graphene.Mutation):
    class Arguments:
        application_id = graphene.ID(required=True)

    ok = graphene.Boolean()
    application = graphene.Field(ApplicationAdminType)

    @classmethod
    @login_required
    @permission_required('auth.human_resources')
    def mutate(cls, root, info, application_id):
        user = info.context.user
        app = Application.objects.get(pk=application_id)
        if not app.reviewer:
            app.reviewer = user
            app.reviewer_character = user.profile.main_character
            app.save()
            notify(app.user, "Application In Progress", message=f"Your application to {app.form.corp} is being reviewed by {app.reviewer_str}")
            ok = True
        else:
            ok = False

        return cls(ok=ok, application=app)


class Mutation:
    hr_create_application = CreateApplicationMutation.Field()
    hr_delete_application = DeleteApplicationMutation.Field()
    hr_add_application_comment = AddApplicationCommentMutation.Field()
    hr_admin_remove_application = AdminRemoveApplicationMutation.Field()
    hr_admin_approve_application = AdminApproveApplication.Field()
    hr_admin_reject_application = AdminRejectApplication.Field()
    hr_admin_mark_in_progress_application = AdminMarkInProgressApplication.Field()



import graphene
from graphql_jwt.decorators import login_required, permission_required

from django.utils import timezone
from django.db.models import Q, Exists, F, OuterRef
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.conf import settings

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter


from allianceauth_pve.models import Rotation, PveButton, RoleSetup, General
from allianceauth_pve.actions import running_averages
from allianceauth_graphql.eveonline.types import EveCharacterType

from .types import RotationType, RoleSetupType, RattingSummaryType, PveButtonType


User = get_user_model()


class Query:
    pve_get_rotation = graphene.Field(RotationType, id=graphene.Int(required=True))
    pve_closed_rotations = graphene.List(RotationType)
    pve_char_running_averages = graphene.Field(RattingSummaryType, start_date=graphene.Date(required=True), end_date=graphene.Date())
    pve_active_rotations = graphene.List(RotationType)
    pve_search_rotation_characters = graphene.List(EveCharacterType, name=graphene.String(), exclude_characters_ids=graphene.List(graphene.Int))
    pve_roles_setups = graphene.List(RoleSetupType)
    pve_buttons = graphene.List(PveButtonType)

    @login_required
    @permission_required('allianceauth_pve.access_pve')
    def resolve_pve_get_rotation(self, info, id):
        return Rotation.objects.get(pk=id)

    @login_required
    @permission_required('allianceauth_pve.access_pve')
    def resolve_pve_closed_rotations(self, info):
        return Rotation.objects.filter(is_closed=True).order_by('-closed_at')

    @login_required
    def resolve_pve_char_running_averages(self, info, start_date, end_date=None):
        return running_averages(info.context.user, start_date, end_date)

    @login_required
    @permission_required('allianceauth_pve.access_pve')
    def resolve_pve_active_rotations(self, info):
        return Rotation.objects.filter(is_closed=False).order_by('-priority')

    @login_required
    @permission_required('allianceauth_pve.manage_entries')
    def resolve_pve_search_rotation_characters(self, info, name=None, exclude_characters_ids=[]):
        content_type = ContentType.objects.get_for_model(General)
        permission = Permission.objects.get(
            content_type=content_type,
            codename='access_pve'
        )

        ownerships = CharacterOwnership.objects.filter(
            Q(user__groups__permissions=permission) |
            Q(user__user_permissions=permission) |
            Q(user__profile__state__permissions=permission),
            user__profile__main_character__isnull=False,
        )

        if name:
            alts_name = CharacterOwnership.objects.filter(
                user=OuterRef('user'),
                character__character_name__icontains=name
            )

            ownerships = ownerships.filter(
                Q(character__character_name__icontains=name) |
                (Exists(alts_name) & Q(character=F('user__profile__main_character')))
            )

        if getattr(settings, 'PVE_ONLY_MAINS', False):
            ownerships = ownerships.filter(character=F('user__profile__main_character'))

        results = EveCharacter.objects.filter(pk__in=ownerships.values('character'))

        if len(exclude_characters_ids) > 0:
            results = results.exclude(pk__in=exclude_characters_ids)

        return results

    @login_required
    @permission_required('allianceauth_pve.manage_rotations')
    def resolve_pve_roles_setups(self, info):
        return RoleSetup.objects.all()

    @login_required
    @permission_required('allianceauth_pve.manage_rotations')
    def resolve_pve_buttons(self, info):
        return PveButton.objects.all()



import datetime
from graphene_django.utils.testing import GraphQLTestCase
from django.utils import timezone

from allianceauth.tests.test_auth_utils import AuthUtils
from app_utils.testdata_factories import UserFactory, UserMainFactory

from allianceauth.optimer.models import OpTimer, OpTimerType


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.optimer_view', UserFactory(), False)

        cls.timer_past = OpTimer.objects.create(start=timezone.now() - datetime.timedelta(days=1))
        cls.timer_future = OpTimer.objects.create(start=timezone.now() + datetime.timedelta(days=1))

    def test_optimer_past_timers(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                optimerPastTimers {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerPastTimers': [
                        {
                            'id': str(self.timer_past.pk)
                        }
                    ]
                }
            }
        )

    def test_optimer_future_timers(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                optimerFutureTimers {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerFutureTimers': [
                        {
                            'id': str(self.timer_future.pk)
                        }
                    ]
                }
            }
        )


class TestOpFormMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.optimer_management', UserMainFactory(), False)

        cls.form_data = {
            'doctrine': 'Test Doctrine',
            'system': 'Test System',
            'start': (timezone.now() + datetime.timedelta(days=1)).isoformat(),
            'duration': '1h',
            'operationName': 'Test Operation',
            'fc': 'Test FC',
        }

    def test_ok_no_type(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: OpFormMutationInput!) {
                optimerNewOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerNewOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 0)

    def test_ok_with_existing_type(self):
        self.client.force_login(self.user)

        optimer_tipe = OpTimerType.objects.create(type='Test Type')

        response = self.query(
            '''
            mutation($input: OpFormMutationInput!) {
                optimerNewOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                        type {
                            id
                        }
                    }
                }
            }
            ''',
            input_data={
                **self.form_data,
                'type': optimer_tipe.type
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerNewOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            },
                            'type': {
                                'id': str(optimer_tipe.pk)
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 1)

    def test_ok_with_new_type(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: OpFormMutationInput!) {
                optimerNewOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                        type {
                            type
                        }
                    }
                }
            }
            ''',
            input_data={
                **self.form_data,
                'type': 'Test Type'
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerNewOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            },
                            'type': {
                                'type': 'Test Type'
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 1)


class TestRemoveOpTimerMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.optimer_management', UserMainFactory(), False)

        cls.timer = OpTimer.objects.create()

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($optimerId: ID!) {
                optimerRemoveOp(optimerId: $optimerId) {
                    ok
                }
            }
            ''',
            variables={
                'optimerId': self.timer.pk
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerRemoveOp': {
                        'ok': True
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 0)


class TestOpFormEditMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.optimer_management', UserMainFactory(), False)

        cls.timer = OpTimer.objects.create()

        cls.form_data = {
            'opId': cls.timer.pk,
            'doctrine': 'Test Doctrine',
            'system': 'Test System',
            'start': (timezone.now() + datetime.timedelta(days=1)).isoformat(),
            'duration': '1h',
            'operationName': 'Test Operation',
            'fc': 'Test FC',
        }

    def test_ok_no_type(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: OpFormEditMutationInput!) {
                optimerEditOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerEditOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 0)

    def test_ok_with_existing_type(self):
        optimer_tipe = OpTimerType.objects.create(type='Test Type')

        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: OpFormEditMutationInput!) {
                optimerEditOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                        type {
                            id
                        }
                    }
                }
            }
            ''',
            input_data={
                **self.form_data,
                'type': optimer_tipe.type
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerEditOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            },
                            'type': {
                                'id': str(optimer_tipe.pk)
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 1)

    def test_ok_with_new_type(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: OpFormEditMutationInput!) {
                optimerEditOp(input: $input) {
                    op {
                        doctrine
                        system
                        start
                        duration
                        operationName
                        fc
                        eveCharacter {
                            id
                        }
                        type {
                            type
                        }
                    }
                }
            }
            ''',
            input_data={
                **self.form_data,
                'type': 'Test Type'
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'optimerEditOp': {
                        'op': {
                            'doctrine': self.form_data['doctrine'],
                            'system': self.form_data['system'],
                            'start': self.form_data['start'],
                            'duration': self.form_data['duration'],
                            'operationName': self.form_data['operationName'],
                            'fc': self.form_data['fc'],
                            'eveCharacter': {
                                'id': str(self.user.profile.main_character.pk)
                            },
                            'type': {
                                'type': 'Test Type'
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(OpTimer.objects.count(), 1)
        self.assertEqual(OpTimerType.objects.count(), 1)



import json
import re
from urllib.parse import quote_plus
from unittest.mock import patch
from faker import Faker

from django.test import override_settings, TestCase, modify_settings
from django.core import mail, signing
from django.urls import reverse
from graphene_django.utils.testing import GraphQLTestCase

from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory, UserFactory
from app_utils.testing import add_character_to_user, add_new_token, generate_invalid_pk, create_authgroup

from esi.models import Token
from esi import app_settings
from allianceauth.eveonline.autogroups.models import AutogroupsConfig

from ..authentication.types import LoginStatus


MOCK_REGISTRATION_SALT = "testing"
MOCK_REDIRECT_SITE = 'https://example.com'


@override_settings(REGISTRATION_VERIFY_EMAIL=True)
class TestEsiTokenAuthMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.token = cls.user.token_set.first()

    @patch('esi.managers.TokenManager.create_from_code')
    def test_logged_in(self, mock_create_from_code):
        mock_create_from_code.return_value = self.token

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                    token
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        content = json.loads(response.content)
        token = content['data']['esiTokenAuth']['token']

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': [],
                        'status': LoginStatus.LOGGED_IN.name,
                        'me': {
                            'id': str(self.user.pk)
                        },
                        'token': token
                    }
                }
            }
        )

        response = self.query(
            '''
            query q {
                me {
                    id
                }
            }
            ''',
            operation_name='q',
            headers={'HTTP_AUTHORIZATION': f'JWT {token}'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'me': {
                        'id': str(self.user.pk)
                    }
                }
            }
        )

    @patch('allianceauth_graphql.authentication.mutations.authenticate')
    @patch('esi.managers.TokenManager.create_from_code')
    def test_user_none(self, mock_create_from_code, mock_authenticate):
        mock_create_from_code.return_value = self.token
        mock_authenticate.return_value = None

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': [
                            'Unable to authenticate the selected character',
                        ],
                        'status': LoginStatus.ERROR.name,
                        'me': None
                    }
                }
            }
        )

    @patch('esi.managers.TokenManager.create_from_code')
    def test_email_registration(self, mock_create_from_code):
        mock_create_from_code.return_value = self.token

        self.user.email = ''
        self.user.is_active = False
        self.user.save()

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                    token
                    refreshToken
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': [],
                        'status': LoginStatus.REGISTRATION.name,
                        'me': None,
                        'refreshToken': None,
                        'token': None
                    }
                }
            }
        )

        self.assertIn('registration_uid', self.client.session)
        self.assertEqual(self.client.session['registration_uid'], self.user.pk)

    @override_settings(REGISTRATION_VERIFY_EMAIL=False)
    @patch('esi.managers.TokenManager.create_from_code')
    def test_email_disabled_registration(self, mock_create_from_code):
        mock_create_from_code.return_value = self.token

        self.user.email = ''
        self.user.is_active = False
        self.user.save()

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': [],
                        'status': LoginStatus.LOGGED_IN.name,
                        'me': {
                            'id': str(self.user.pk)
                        }
                    }
                }
            }
        )

        self.user.refresh_from_db()
        self.assertTrue(self.user.is_active)

    @patch('esi.managers.TokenManager.create_from_code')
    def test_token_equivalent_exists(self, mock_create_from_code):
        newtoken = add_new_token(self.user, self.user.profile.main_character, "publicData")

        mock_create_from_code.return_value = newtoken

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': [],
                        'status': LoginStatus.LOGGED_IN.name,
                        'me': {
                            'id': str(self.user.pk)
                        }
                    }
                }
            }
        )

        self.assertFalse(
            Token.objects
            .filter(pk=newtoken.pk)
            .exists()
        )

    @patch('esi.managers.TokenManager.create_from_code')
    def test_not_able_to_authenticate(self, mock_create_from_code):
        mock_create_from_code.return_value = self.token

        self.user.is_active = False
        self.user.save()

        response = self.query(
            '''
            mutation testM($sso_token: String!) {
                esiTokenAuth(ssoToken: $sso_token) {
                    me {
                        id
                    }
                    errors
                    status
                }
            }
            ''',
            operation_name='testM',
            variables={'sso_token': 'nice_token'}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiTokenAuth': {
                        'errors': ['Unable to authenticate the selected character'],
                        'status': LoginStatus.ERROR.name,
                        'me': None
                    }
                }
            }
        )


class TestRegistrationMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(
            is_active=False,
            email='',
        )

    @override_settings(REDIRECT_SITE='https://www.example.com')
    def test_email_registration(self):
        session = self.client.session
        session.update({'registration_uid': self.user.pk})
        session.save()

        fake = Faker()

        response = self.query(
            '''
            mutation testM($input: RegistrationMutationInput!) {
                authenticationEmailRegistration(input: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data={'email': fake.email()}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationEmailRegistration': {
                        'errors': [],
                        'ok': True
                    }
                }
            }
        )

        self.assertEqual(len(mail.outbox), 1)

    @override_settings(REDIRECT_SITE='https://www.example.com')
    def test_missing_user_id(self):
        fake = Faker()

        response = self.query(
            '''
            mutation testM($input: RegistrationMutationInput!) {
                authenticationEmailRegistration(input: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data={'email': fake.email()}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationEmailRegistration': {
                        'errors': ['You need to do the token registration step first!'],
                        'ok': False
                    }
                }
            }
        )

        self.assertEqual(len(mail.outbox), 0)

    def test_missing_site(self):
        session = self.client.session
        session.update({'registration_uid': self.user.pk})
        session.save()

        fake = Faker()

        response = self.query(
            '''
            mutation testM($input: RegistrationMutationInput!) {
                authenticationEmailRegistration(input: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data={'email': fake.email()}
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationEmailRegistration': {
                        'errors': ['Redirect site not specified in settings!'],
                        'ok': False
                    }
                }
            }
        )

        self.assertEqual(len(mail.outbox), 0)


class TestChangeMainCharacterMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user1, cls.user2 = UserFactory.create_batch(2)

        cls.newchar = EveCharacterFactory()

    def test_ok(self):
        add_character_to_user(self.user1, self.newchar)

        self.client.force_login(self.user1, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                authenticationChangeMainCharacter(newMainCharacterId: $input) {
                    errors
                    ok
                    me {
                        id
                        profile {
                            mainCharacter {
                                id
                            }
                        }
                    }
                }
            }
            ''',
            operation_name='testM',
            input_data=self.newchar.character_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationChangeMainCharacter': {
                        'errors': [],
                        'ok': True,
                        'me': {
                            'id': str(self.user1.pk),
                            'profile': {
                                'mainCharacter': {
                                    'id': str(self.newchar.pk)
                                }
                            }
                        }
                    }
                }
            }
        )

    def test_character_not_added(self):
        self.client.force_login(self.user1, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                authenticationChangeMainCharacter(newMainCharacterId: $input) {
                    errors
                    ok
                    me {
                        id
                        profile {
                            mainCharacter {
                                id
                            }
                        }
                    }
                }
            }
            ''',
            operation_name='testM',
            input_data=self.newchar.character_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationChangeMainCharacter': {
                        'errors': ["You never added this character"],
                        'ok': False,
                        'me': {
                            'id': str(self.user1.pk),
                            'profile': {
                                'mainCharacter': None
                            }
                        }
                    }
                }
            }
        )

    def test_character_not_owned(self):
        add_character_to_user(self.user2, self.newchar)

        self.client.force_login(self.user1, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                authenticationChangeMainCharacter(newMainCharacterId: $input) {
                    errors
                    ok
                    me {
                        id
                        profile {
                            mainCharacter {
                                id
                            }
                        }
                    }
                }
            }
            ''',
            operation_name='testM',
            input_data=self.newchar.character_id
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationChangeMainCharacter': {
                        'errors': ["You don't own this character"],
                        'ok': False,
                        'me': {
                            'id': str(self.user1.pk),
                            'profile': {
                                'mainCharacter': None
                            }
                        }
                    }
                }
            }
        )


class TestAddCharacterMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()
        cls.newchar = EveCharacterFactory()

    @patch('esi.managers.TokenManager.create_from_code')
    def test_ok(self, mock_create_from_code):
        mock_create_from_code.return_value = add_new_token(self.user, self.newchar)

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: String!) {
                authenticationAddCharacter(newCharSsoToken: $input) {
                    errors
                    ok
                    me {
                        id
                        characterOwnerships {
                            character {
                                id
                            }
                        }
                    }
                }
            }
            ''',
            operation_name='testM',
            input_data='nice_token'
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationAddCharacter': {
                        'errors': [],
                        'ok': True,
                        'me': {
                            'id': str(self.user.pk),
                            'characterOwnerships': [
                                {
                                    'character': {
                                        'id': str(self.newchar.pk)
                                    }
                                }
                            ]
                        }
                    }
                }
            }
        )
    


class TestRefreshEsiTokenMutation(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()

    @patch('esi.models.Token.refresh')
    def test_ok(self, mock_refresh):
        mock_refresh.return_value = None

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        newtoken = add_new_token(
            self.user,
            EveCharacterFactory()
        )

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRefreshToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=newtoken.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRefreshToken': {
                        'errors': [],
                        'ok': True,
                    }
                }
            }
        )

        self.assertTrue(mock_refresh.called)

    def test_token_not_exists(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRefreshToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=generate_invalid_pk(Token)
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRefreshToken': {
                        'errors': ["Token does not exist"],
                        'ok': False,
                    }
                }
            }
        )

    def test_token_not_belongs_to_user(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        newuser = UserFactory()
        newtoken = add_new_token(
            newuser,
            EveCharacterFactory()
        )

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRefreshToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=newtoken.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRefreshToken': {
                        'errors': ["This token does not belong to you."],
                        'ok': False,
                    }
                }
            }
        )

    @patch('esi.models.Token.refresh')
    def test_exception(self, mock_refresh):
        mock_refresh.side_effect = Exception('Test')

        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        newtoken = add_new_token(
            self.user,
            EveCharacterFactory()
        )

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRefreshToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=newtoken.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRefreshToken': {
                        'errors': ["Failed to refresh token. Test"],
                        'ok': False,
                    }
                }
            }
        )

        self.assertTrue(mock_refresh.called)


class TestRemoveEsiTokenMutation(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.token = cls.user.token_set.first()

    def test_ok(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRemoveToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=self.token.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRemoveToken': {
                        'errors': [],
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(self.user.token_set.count(), 0)

    def test_token_not_belongs_to_user(self):
        user2 = UserFactory()

        self.client.force_login(user2, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRemoveToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=self.token.pk
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRemoveToken': {
                        'errors': ["This token does not belong to you."],
                        'ok': False,
                    }
                }
            }
        )

        self.assertEqual(self.user.token_set.count(), 1)

    def test_token_not_exists(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            mutation testM($input: Int!) {
                esiRemoveToken(tokenId: $input) {
                    errors
                    ok
                }
            }
            ''',
            operation_name='testM',
            input_data=generate_invalid_pk(Token)
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiRemoveToken': {
                        'errors': ["Token does not exist"],
                        'ok': False,
                    }
                }
            }
        )

        self.assertEqual(self.user.token_set.count(), 1)


@override_settings(
    REGISTRATION_SALT=MOCK_REGISTRATION_SALT,
    REDIRECT_SITE=MOCK_REDIRECT_SITE
)
class TestVerifyEmailView(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory(is_active=False)

    def test_ok(self):
        activation_key = signing.dumps([self.user.pk, self.user.email], salt=MOCK_REGISTRATION_SALT)

        email = self.user.email
        self.user.email = ''
        self.user.save()

        response = self.client.get(reverse('allianceauth_graphql:verify_email') + f"?activation_key={activation_key}")

        self.assertRedirects(
            response,
            MOCK_REDIRECT_SITE + '/registration/callback/',
            fetch_redirect_response=False
        )

        self.user.refresh_from_db()
        self.assertEqual(self.user.email, email)
        self.assertTrue(self.user.is_active)

    def test_not_ok(self):
        activation_key = signing.dumps([self.user.pk, self.user.email], salt=MOCK_REGISTRATION_SALT + 'WRONG')

        self.user.email = ''
        self.user.save()

        response = self.client.get(reverse('allianceauth_graphql:verify_email') + f"?activation_key={activation_key}")

        self.assertContains(response, 'Invalid signature', html=True)

        self.user.refresh_from_db()
        self.assertEqual(self.user.email, '')
        self.assertFalse(self.user.is_active)


class TestQueries(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_login_url(self):
        response = self.query(
            '''
            query q {
                esiLoginUrl(scopes: ["test1", "test2"])
            }
            '''
        )

        data = json.loads(response.content)
        self.assertIn('data', data)
        self.assertIn('esiLoginUrl', data['data'])

        login_url = data['data']['esiLoginUrl']

        self.assertRegex(login_url, rf"{re.escape(app_settings.ESI_OAUTH_LOGIN_URL)}\?response_type=code\&client_id=[0-9a-z]+\&redirect_uri={re.escape(quote_plus(app_settings.ESI_SSO_CALLBACK_URL))}\&scope={re.escape(quote_plus('test1 test2'))}\&state=[0-9a-zA-Z]+")

    def test_me(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                me {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'me': {
                        'id': str(self.user.pk)
                    }
                }
            }
        )

    @modify_settings(INSTALLED_APPS={'remove': ['allianceauth.eveonline.autogroups']})
    def test_user_groups(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        group1 = create_authgroup()
        self.user.groups.add(group1)

        response = self.query(
            '''
            query q {
                authenticationUserGroups {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationUserGroups': [
                        {
                            'id': str(group1.pk)
                        }
                    ]
                }
            }
        )

    def test_user_groups_exclude_autogroups(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        group1 = create_authgroup()
        self.user.groups.add(group1)

        config: AutogroupsConfig = AutogroupsConfig.objects.create(
            corp_groups=True,
            alliance_groups=True
        )

        config.states.add(self.user.profile.state)

        response = self.query(
            '''
            query q {
                authenticationUserGroups {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationUserGroups': [
                        {
                            'id': str(group1.pk)
                        }
                    ]
                }
            }
        )

    def test_user_characters(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                authenticationUserCharacters {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'authenticationUserCharacters': [
                        {
                            'id': str(self.user.profile.main_character.pk)
                        }
                    ]
                }
            }
        )



from datetime import timedelta

from graphene_django.utils.testing import GraphQLTestCase
from unittest.mock import patch

from django.utils import timezone
from django.core.exceptions import ValidationError

from allianceauth.tests.test_auth_utils import AuthUtils
from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory, UserFactory
from app_utils.testing import add_character_to_user, generate_invalid_pk
from app_utils.esi_testing import EsiEndpoint, EsiClientStub

from allianceauth.fleetactivitytracking.models import Fatlink, Fat
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.models import Token


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'auth.fleetactivitytracking_statistics',
                'auth.fleetactivitytracking',
            ],
            cls.user,
            False
        )

        cls.corp: EveCorporationInfo = cls.user.profile.main_character.corporation

        cls.extrachar1, *charlist = EveCharacterFactory.create_batch(7, corporation=cls.corp)

        cls.user2 = UserMainFactory(main_character__character=cls.extrachar1)

        for char in charlist:
            add_character_to_user(cls.user2, char)

        cls.fatlink = Fatlink.objects.create(
            creator=cls.user,
            fatdatetime=timezone.now(),
            fleet='Test Fatlink',
            hash='testhash',
            duration=60,
        )

        for char in EveCharacter.objects.select_related('character_ownership__user').all():
            Fat.objects.create(
                character=char,
                fatlink=cls.fatlink,
                shiptype='Test Ship',
                system='Test System',
                station='Test Station',
                user=char.character_ownership.user,
            )

    def test_recent_fats(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                fatRecentFat {
                    id
                }
            }
            ''',
        )

        fat = Fat.objects.get(user=self.user)

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatRecentFat': [
                        {
                            'id': str(fat.id),
                        }
                    ]
                }
            }
        )

    def test_fatlinks(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                fatGetFatlinks {
                    id
                }
            }
            ''',
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatGetFatlinks': [
                        {
                            'id': str(self.fatlink.id),
                        }
                    ]
                }
            }
        )

    def test_fat_corp_monthly_stats(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($year: Int!, $month: Int!, $corpId: Int!) {
                fatCorpMonthlyStats(year: $year, month: $month, corpId: $corpId) {
                    user {
                        id
                    }
                    numChars
                }
            }
            ''',
            variables={
                'year': timezone.now().year,
                'month': timezone.now().month,
                'corpId': self.corp.corporation_id
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatCorpMonthlyStats': [
                        {
                            'user': {
                                'id': str(self.user2.id),
                            },
                            'numChars': 7,
                        },
                        {
                            'user': {
                                'id': str(self.user.id),
                            },
                            'numChars': 1,
                        },
                    ]
                }
            }
        )

    def test_fat_general_monthly_stats(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($year: Int!, $month: Int!) {
                fatGeneralMonthlyStats(year: $year, month: $month) {
                    corporation {
                        id
                    }
                    numFats
                }
            }
            ''',
            variables={
                'year': timezone.now().year,
                'month': timezone.now().month,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatGeneralMonthlyStats': [
                        {
                            'corporation': {
                                'id': str(self.corp.id),
                            },
                            'numFats': 8,
                        }
                    ]
                }
            }
        )

    def test_fat_personal_stats(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                fatPersonalStats {
                    month
                    year
                    numFats
                }
            }
            ''',
        )

        fat = Fat.objects.select_related('fatlink').get(user=self.user)

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatPersonalStats': [
                        {
                            'month': fat.fatlink.fatdatetime.month,
                            'year': fat.fatlink.fatdatetime.year,
                            'numFats': 1,
                        }
                    ]
                }
            }
        )

    def test_fat_personal_monthly_stats_self(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($year: Int!, $month: Int!, $charId: Int) {
                fatPersonalMonthlyStats(year: $year, month: $month, charId: $charId) {
                    collectedLinks {
                        shiptype
                        timesUsed
                    }
                    createdLinks {
                        id
                    }
                }
            }
            ''',
            variables={
                'year': timezone.now().year,
                'month': timezone.now().month,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatPersonalMonthlyStats': {
                        'collectedLinks': [
                            {
                                'shiptype': 'Test Ship',
                                'timesUsed': 1,
                            }
                        ],
                        'createdLinks': [
                            {
                                'id': str(self.fatlink.id),
                            }
                        ]
                    }
                }
            }
        )

    def test_fat_personal_monthly_stats_other(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($year: Int!, $month: Int!, $charId: Int) {
                fatPersonalMonthlyStats(year: $year, month: $month, charId: $charId) {
                    collectedLinks {
                        shiptype
                        timesUsed
                    }
                    createdLinks {
                        id
                    }
                }
            }
            ''',
            variables={
                'year': timezone.now().year,
                'month': timezone.now().month,
                'charId': self.extrachar1.character_id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatPersonalMonthlyStats': {
                        'collectedLinks': [
                            {
                                'shiptype': 'Test Ship',
                                'timesUsed': 7,
                            }
                        ],
                        'createdLinks': []
                    }
                }
            }
        )


class TestAddFatParticipationMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.scopes = [
            'esi-location.read_location.v1',
            'esi-location.read_ship_type.v1',
            'esi-universe.read_structures.v1',
            'esi-location.read_online.v1',
        ]

        cls.user = UserMainFactory(
            main_character__scopes=cls.scopes
        )
        cls.token1 = cls.user.token_set.first()
        cls.mainchar = cls.user.profile.main_character

        cls.char2 = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char2, scopes=cls.scopes)
        cls.token2 = cls.user.token_set.get(character_id=cls.char2.character_id)

        cls.fatlink = Fatlink.objects.create(
            creator=cls.user,
            fatdatetime=timezone.now(),
            fleet='Test Fatlink',
            hash='testhash',
            duration=60,
        )

        cls.endpoints = [
            EsiEndpoint(
                'Location',
                'get_characters_character_id_online',
                'character_id',
            ),
            EsiEndpoint(
                'Location',
                'get_characters_character_id_location',
                'character_id',
            ),
            EsiEndpoint(
                'Location',
                'get_characters_character_id_ship',
                'character_id',
            ),
            EsiEndpoint(
                'Universe',
                'get_universe_systems_system_id',
                'system_id',
            ),
            EsiEndpoint(
                'Universe',
                'get_universe_stations_station_id',
                'station_id',
            ),
            EsiEndpoint(
                'Universe',
                'get_universe_structures_structure_id',
                'structure_id',
            )
        ]

        cls.data = {
            'Location': {
                'get_characters_character_id_online': {
                    str(cls.mainchar.character_id): {
                        'online': True,
                    },
                    str(cls.char2.character_id): {
                        'online': False,
                    },
                },
                'get_characters_character_id_location': {
                    str(cls.mainchar.character_id): {
                        'solar_system_id': 30000142,
                        'station_id': None,
                        'structure_id': None,
                    }
                },
                'get_characters_character_id_ship': {
                    str(cls.mainchar.character_id): {
                        'ship_type_id': 123,
                    }
                }
            },
            'Universe': {
                'get_universe_systems_system_id': {
                    '30000142': {
                        'name': 'Jita',
                    }
                },
                'get_universe_stations_station_id': {
                    '123456789': {
                        'name': 'Test Station',
                    }
                },
                'get_universe_structures_structure_id': {
                    '1234567890': {
                        'name': 'Test Structure',
                    }
                }
            }
        }

    def test_token_missing(self):
        user2 = UserMainFactory()
        self.client.force_login(user2)

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': None
                },
                'errors': [
                    {
                        'message': 'Required token missing',
                        'locations': [
                            {
                                'line': 3,
                                'column': 17
                            }
                        ],
                        'path': [
                            'fatParticipateToFatlink'
                        ]
                    }
                ]
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    @patch('allianceauth_graphql.fleetactivitytracking.mutations.provider.get_itemtype')
    @patch('esi.models.Token.get_esi_client')
    def test_ok(self, mock_get_esi_client, mock_get_itemtype):
        self.client.force_login(self.user)

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)
        mock_get_itemtype.return_value = type('Item', (object,), {'name': 'Test Ship'})

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': True,
                        'error': None,
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 1)

        fat = Fat.objects.first()

        self.assertEqual(fat.station, "No Station")

    def test_invalid_token(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': generate_invalid_pk(Token),
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': False,
                        'error': 'Token not valid',
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    @patch('esi.models.Token.get_esi_client')
    def test_character_not_online(self, mock_get_esi_client):
        self.client.force_login(self.user)

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token2.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': False,
                        'error': f"Cannot register the fleet participation for {self.char2.character_name}. The character needs to be online.",
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    @patch('esi.models.Token.get_esi_client')
    def test_fatlink_expired(self, mock_get_esi_client):
        self.client.force_login(self.user)

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)

        self.fatlink.fatdatetime = timezone.now() - timedelta(minutes=self.fatlink.duration * 2)
        self.fatlink.save()

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': False,
                        'error': 'FAT link has expired or user not valid',
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    @patch('esi.models.Token.get_esi_client')
    def test_character_not_exists(self, mock_get_esi_client):
        self.client.force_login(self.user)

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)

        self.mainchar.delete()

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': False,
                        'error': "Character doesn't exists",
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    @patch('allianceauth_graphql.fleetactivitytracking.mutations.provider.get_itemtype')
    @patch('esi.models.Token.get_esi_client')
    def test_in_station(self, mock_get_esi_client, mock_get_itemtype):
        self.client.force_login(self.user)

        self.data['Location']['get_characters_character_id_location'][str(self.mainchar.character_id)]['station_id'] = 123456789

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)
        mock_get_itemtype.return_value = type('Item', (object,), {'name': 'Test Ship'})

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': True,
                        'error': None,
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 1)

        fat = Fat.objects.first()

        self.assertEqual(fat.station, 'Test Station')

    @patch('allianceauth_graphql.fleetactivitytracking.mutations.provider.get_itemtype')
    @patch('esi.models.Token.get_esi_client')
    def test_in_structure(self, mock_get_esi_client, mock_get_itemtype):
        self.client.force_login(self.user)

        self.data['Location']['get_characters_character_id_location'][str(self.mainchar.character_id)]['structure_id'] = 1234567890

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)
        mock_get_itemtype.return_value = type('Item', (object,), {'name': 'Test Ship'})

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': True,
                        'error': None,
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 1)

        fat = Fat.objects.first()

        self.assertEqual(fat.station, 'Test Structure')

    @patch('allianceauth.fleetactivitytracking.models.Fat.full_clean')
    @patch('allianceauth_graphql.fleetactivitytracking.mutations.provider.get_itemtype')
    @patch('esi.models.Token.get_esi_client')
    def test_validation_error(self, mock_get_esi_client, mock_get_itemtype, mock_full_clean):
        self.client.force_login(self.user)

        mock_get_esi_client.return_value = EsiClientStub(self.data, self.endpoints)
        mock_get_itemtype.return_value = type('Item', (object,), {'name': 'Test Ship'})
        mock_full_clean.side_effect = ValidationError({'test': 'Test Error'})

        response = self.query(
            '''
            mutation($tokenId: ID!, $fatlinkHash: String!) {
                fatParticipateToFatlink(tokenId: $tokenId, fatlinkHash: $fatlinkHash) {
                    ok
                    error
                }
            }
            ''',
            variables={
                'tokenId': self.token1.id,
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatParticipateToFatlink': {
                        'ok': False,
                        'error': 'Test Error',
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)


class TestCreateFatlinkMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.fleetactivitytracking', UserFactory(), False)

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: CreateFatlinkInput!) {
                fatCreateFatlink(input: $input) {
                    ok
                }
            }
            ''',
            input_data={
                'fleet': 'Test Fleet',
                'duration': 60,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatCreateFatlink': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Fatlink.objects.count(), 1)

        fatlink = Fatlink.objects.first()

        self.assertEqual(fatlink.fleet, 'Test Fleet')
        self.assertEqual(fatlink.duration, 60)


class TestOtherMutations(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.fleetactivitytracking', UserMainFactory(), False)

        cls.mainchar = cls.user.profile.main_character

        cls.fatlink = Fatlink.objects.create(
            creator=cls.user,
            fatdatetime=timezone.now(),
            fleet='Test Fatlink',
            hash='testhash',
            duration=60,
        )

        cls.fat = Fat.objects.create(
            fatlink=cls.fatlink,
            character=cls.mainchar,
            shiptype='Test Ship',
            system='Test System',
            station='Test Station',
            user=cls.user,
        )

    def test_remove_char_fatlink(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fatlinkHash: String!, $characterId: Int!) {
                fatRemoveCharFat(fatlinkHash: $fatlinkHash, characterId: $characterId) {
                    ok
                }
            }
            ''',
            variables={
                'fatlinkHash': self.fatlink.hash,
                'characterId': self.mainchar.character_id,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatRemoveCharFat': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Fat.objects.count(), 0)

    def test_delete_fatlink(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($fatlinkHash: String!) {
                fatDeleteFatlink(fatlinkHash: $fatlinkHash) {
                    ok
                }
            }
            ''',
            variables={
                'fatlinkHash': self.fatlink.hash,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'fatDeleteFatlink': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Fatlink.objects.count(), 0)



import graphene


class ApplicationResponseInputType(graphene.InputObjectType):
    question_id = graphene.Int(required=True)
    answer = graphene.List(graphene.String, required=True)


class FormAnswerInputType(graphene.InputObjectType):
    responses = graphene.List(ApplicationResponseInputType, required=True)
    form_id = graphene.ID(required=True)



import datetime
from graphene_django.utils.testing import GraphQLTestCase

from django.utils import timezone
from django.test import override_settings

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testdata_factories import UserMainFactory, EveCharacterFactory, UserFactory
from app_utils.testing import add_character_to_user, generate_invalid_pk

from allianceauth_pve.models import Rotation, Entry, EntryCharacter, EntryRole, PveButton

from ..community_creations.allianceauth_pve_integration.inputs import EntryInput


class TestQueries(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_entries',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.char2 = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char2)

        cls.open_rotation: Rotation = Rotation.objects.create(name='Open Rotation')

        cls.closed_rotation: Rotation = Rotation.objects.create(
            name='Closed Rotation',
            is_closed=True,
            actual_total=100,
            closed_at=timezone.now() - datetime.timedelta(days=1),
        )

        cls.entry = Entry.objects.create(
            rotation=cls.closed_rotation,
            estimated_total=100,
            created_by=cls.user
        )

        role = EntryRole.objects.create(
            entry=cls.entry,
            name='Role',
            value=1
        )

        EntryCharacter.objects.create(
            entry=cls.entry,
            user=cls.user,
            user_character=cls.mainchar,
            role=role,
        )

    def test_pve_get_rotation(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($id: Int!) {
                pveGetRotation(id: $id) {
                    id
                    entries {
                        id
                    }
                    summary {
                        mainCharacter {
                            id
                        }
                    }
                }
            }
            ''',
            variables={'id': self.closed_rotation.pk},
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveGetRotation': {
                        'id': str(self.closed_rotation.pk),
                        'entries': [
                            {
                                'id': str(self.entry.pk),
                            }
                        ],
                        'summary': [
                            {
                                'mainCharacter': {
                                    'id': str(self.mainchar.pk),
                                }
                            }
                        ]
                    }
                }
            }
        )

    def test_pve_closed_rotations(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                pveClosedRotations {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveClosedRotations': [
                        {
                            'id': str(self.closed_rotation.pk),
                        }
                    ]
                }
            }
        )

    def test_pve_char_running_averages(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($start_date: Date!, $end_date: Date) {
                pveCharRunningAverages(startDate: $start_date, endDate: $end_date) {
                    actualTotal
                    estimatedTotal
                    mainCharacter {
                        id
                    }
                }
            }
            ''',
            variables={'start_date': '2020-01-01'},
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCharRunningAverages': {
                        'actualTotal': 100.0,
                        'estimatedTotal': 100.0,
                        'mainCharacter': None,
                    },
                }
            }
        )

    def test_pve_active_rotations(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                pveActiveRotations {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveActiveRotations': [
                        {
                            'id': str(self.open_rotation.pk),
                        }
                    ]
                }
            }
        )

    def test_pve_search_rotation_characters_all(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($name: String, $excludeCharactersIds: [Int!]) {
                pveSearchRotationCharacters(name: $name, excludeCharactersIds: $excludeCharactersIds) {
                    id
                }
            }
            ''',
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveSearchRotationCharacters': [
                        {
                            'id': str(self.mainchar.pk),
                        },
                        {
                            'id': str(self.char2.pk),
                        }
                    ]
                }
            }
        )

    def test_pve_search_rotation_characters_name(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($name: String, $excludeCharactersIds: [Int!]) {
                pveSearchRotationCharacters(name: $name, excludeCharactersIds: $excludeCharactersIds) {
                    id
                }
            }
            ''',
            variables={
                'name': self.char2.character_name
            },
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveSearchRotationCharacters': [
                        {
                            'id': str(self.mainchar.pk),
                        },
                        {
                            'id': str(self.char2.pk),
                        },
                    ]
                }
            }
        )

    @override_settings(PVE_ONLY_MAINS=True)
    def test_pve_search_rotation_characters_name_only_mains(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($name: String, $excludeCharactersIds: [Int!]) {
                pveSearchRotationCharacters(name: $name, excludeCharactersIds: $excludeCharactersIds) {
                    id
                }
            }
            ''',
            variables={'name': self.char2.character_name},
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveSearchRotationCharacters': [
                        {
                            'id': str(self.mainchar.pk),
                        }
                    ]
                }
            }
        )

    def test_pve_search_rotation_characters_exclude(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($name: String, $excludeCharactersIds: [Int!]) {
                pveSearchRotationCharacters(name: $name, excludeCharactersIds: $excludeCharactersIds) {
                    id
                }
            }
            ''',
            variables={'excludeCharactersIds': [self.mainchar.pk]},
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveSearchRotationCharacters': [
                        {
                            'id': str(self.char2.pk),
                        }
                    ]
                }
            }
        )

    def test_pve_roles_setups(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                pveRolesSetups {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveRolesSetups': []
                }
            }
        )

    def test_pve_buttons(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                pveButtons {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveButtons': [
                        {
                            'id': str(button.pk),
                        } for button in PveButton.objects.all()
                    ]
                }
            }
        )


class TestEntryInput(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()
        cls.mainchar = cls.user.profile.main_character

        cls.user2 = UserMainFactory()
        cls.mainchar2 = cls.user2.profile.main_character

        cls.input_data = {
            'roles': [
                {
                    'name': 'role1',
                    'value': 1,
                }
            ],
            'shares': [
                {
                    'character_id': cls.mainchar.pk,
                    'user_id': cls.user.pk,
                    'site_count': 1,
                    'role': 'role1',
                },
                {
                    'character_id': cls.mainchar2.pk,
                    'user_id': cls.user2.pk,
                    'site_count': 1,
                    'role': 'role1',
                }
            ]
        }

    def test_ok(self):
        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))
        self.assertEqual(errors, [])

    def test_duplicated_roles(self):
        self.input_data['roles'].append(self.input_data['roles'][0])

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, [f"{self.input_data['roles'][0]['name']} name is not unique"])

    def test_duplicated_character(self):
        self.input_data['shares'].append(self.input_data['shares'][0])

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, [f"character {self.input_data['shares'][0]['character_id']} cannot have more than 1 share"])

    def test_invalid_role(self):
        self.input_data['shares'][0]['role'] = 'invalid'

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, [f"{self.input_data['shares'][0]['role']} is not a valid role"])

    def test_wrong_ownership(self):
        self.input_data['shares'][0]['user_id'] = self.user2.pk

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, ["character ownership doesn't match"])

    def test_invalid_total(self):
        self.input_data['shares'][0]['site_count'] = 0
        self.input_data['shares'][1]['site_count'] = 0

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, ["Form not valid, you need at least 1 person to receive loot"])

    def test_no_roles_or_shares(self):
        self.input_data['roles'] = []
        self.input_data['shares'] = []

        errors = EntryInput.is_valid(type('EntryInput', (object,), self.input_data))

        self.assertEqual(errors, ["Form not valid, you need at least 1 person to receive loot", "Not enough shares or roles"])


class TestCreateRattingEntry(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_entries',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.char2 = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char2)

        cls.rotation: Rotation = Rotation.objects.create(name='Open Rotation')

        cls.input_data = {
            'estimatedTotal': 1000,
            'roles': [
                {
                    'name': 'role2',
                    'value': 1,
                }
            ],
            'shares': [
                {
                    'characterId': cls.mainchar.pk,
                    'userId': cls.user.pk,
                    'siteCount': 1,
                    'role': 'role2',
                },
                {
                    'characterId': cls.char2.pk,
                    'userId': cls.user.pk,
                    'siteCount': 1,
                    'role': 'role2',
                    'helpedSetup': True,
                }
            ]
        }

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: EntryInput!, $rotationId: Int!) {
                pveCreateEntry(input: $input, rotationId: $rotationId) {
                    ok
                    errors
                }
            }
            ''',
            variables={
                'input': self.input_data,
                'rotationId': self.rotation.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCreateEntry': {
                        'ok': True,
                        'errors': [],
                    }
                }
            }
        )

        self.assertEqual(Entry.objects.count(), 1)

    def test_invalid_rotation(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: EntryInput!, $rotationId: Int!) {
                pveCreateEntry(input: $input, rotationId: $rotationId) {
                    ok
                    errors
                }
            }
            ''',
            variables={
                'input': self.input_data,
                'rotationId': generate_invalid_pk(Rotation),
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCreateEntry': {
                        'ok': False,
                        'errors': ["Rotation doesn't exists"],
                    }
                }
            }
        )

        self.assertEqual(Entry.objects.count(), 0)


class TestModifyRattingEntry(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_entries',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.char2 = EveCharacterFactory()
        add_character_to_user(cls.user, cls.char2)

        cls.user2 = UserFactory()

        cls.rotation: Rotation = Rotation.objects.create(name='Open Rotation')

        cls.entry = Entry.objects.create(
            rotation=cls.rotation,
            estimated_total=100,
            created_by=cls.user
        )

        role = EntryRole.objects.create(
            entry=cls.entry,
            name='Role',
            value=1
        )

        EntryCharacter.objects.create(
            entry=cls.entry,
            user=cls.user,
            user_character=cls.mainchar,
            role=role,
        )

        cls.input_data = {
            'estimatedTotal': 1000,
            'roles': [
                {
                    'name': 'role2',
                    'value': 1,
                }
            ],
            'shares': [
                {
                    'characterId': cls.mainchar.pk,
                    'userId': cls.user.pk,
                    'siteCount': 1,
                    'role': 'role2',
                },
                {
                    'characterId': cls.char2.pk,
                    'userId': cls.user.pk,
                    'siteCount': 1,
                    'role': 'role2',
                    'helpedSetup': True,
                }
            ]
        }

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: EntryInput!, $entryId: Int!) {
                pveModifyEntry(input: $input, entryId: $entryId) {
                    ok
                    errors
                    entry {
                        id
                        estimatedTotal
                    }
                }
            }
            ''',
            variables={
                'input': self.input_data,
                'entryId': self.entry.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveModifyEntry': {
                        'ok': True,
                        'errors': [],
                        'entry': {
                            'id': str(self.entry.pk),
                            'estimatedTotal': 1000,
                        }
                    }
                }
            }
        )

        self.assertEqual(Entry.objects.count(), 1)
        self.assertEqual(self.entry.ratting_shares.count(), 2)

    def test_cannot_edit(self):
        self.client.force_login(self.user)

        self.rotation.is_closed = True
        self.rotation.save()

        response = self.query(
            '''
            mutation($input: EntryInput!, $entryId: Int!) {
                pveModifyEntry(input: $input, entryId: $entryId) {
                    ok
                    errors
                }
            }
            ''',
            variables={
                'input': self.input_data,
                'entryId': self.entry.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveModifyEntry': {
                        'ok': False,
                        'errors': ["You cannot edit this entry"],
                    }
                }
            }
        )


class TestDeleteRattingEntry(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_entries',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.user2 = UserFactory()

        cls.rotation: Rotation = Rotation.objects.create(name='Open Rotation')

        cls.entry = Entry.objects.create(
            rotation=cls.rotation,
            estimated_total=100,
            created_by=cls.user
        )

        role = EntryRole.objects.create(
            entry=cls.entry,
            name='Role',
            value=1
        )

        EntryCharacter.objects.create(
            entry=cls.entry,
            user=cls.user,
            user_character=cls.mainchar,
            role=role,
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($entryId: Int!) {
                pveDeleteEntry(entryId: $entryId) {
                    ok
                }
            }
            ''',
            variables={
                'entryId': self.entry.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveDeleteEntry': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Entry.objects.count(), 0)

    def test_cannot_delete(self):
        self.client.force_login(self.user)

        self.rotation.is_closed = True
        self.rotation.save()

        response = self.query(
            '''
            mutation($entryId: Int!) {
                pveDeleteEntry(entryId: $entryId) {
                    ok
                }
            }
            ''',
            variables={
                'entryId': self.entry.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveDeleteEntry': {
                        'ok': False,
                    }
                }
            }
        )

        self.assertEqual(Entry.objects.count(), 1)


class TestCreateRotation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.user2 = UserFactory()

        cls.input_data = {
            'name': 'Test Rotation',
            'priority': 1,
            'taxRate': 10.0,
            'maxDailySetups': 1,
            'minPeopleShareSetup': 3
        }

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: CreateRotationInput!) {
                pveCreateRotation(input: $input) {
                    rotation {
                        name
                    }
                }
            }
            ''',
            variables={
                'input': self.input_data,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCreateRotation': {
                        'rotation': {
                            'name': 'Test Rotation',
                        }
                    }
                }
            }
        )

        self.assertEqual(Rotation.objects.count(), 1)


class TestCloseRotation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'allianceauth_pve.access_pve',
                'allianceauth_pve.manage_rotations',
            ],
            UserMainFactory(),
            False
        )

        cls.mainchar = cls.user.profile.main_character

        cls.user2 = UserFactory()

        cls.rotation: Rotation = Rotation.objects.create(name='Open Rotation')

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: RotationCloseInput!) {
                pveCloseRotation(input: $input) {
                    ok
                }
            }
            ''',
            input_data={
                'rotationId': self.rotation.pk,
                'salesValue': 0,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCloseRotation': {
                        'ok': True,
                    }
                }
            }
        )

        self.rotation.refresh_from_db()
        self.assertTrue(self.rotation.is_closed)

    def test_already_closed(self):
        self.client.force_login(self.user)

        self.rotation.is_closed = True
        self.rotation.save()

        response = self.query(
            '''
            mutation($input: RotationCloseInput!) {
                pveCloseRotation(input: $input) {
                    ok
                }
            }
            ''',
            input_data={
                'rotationId': self.rotation.pk,
                'salesValue': 0,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'pveCloseRotation': {
                        'ok': False,
                    }
                }
            }
        )

        self.rotation.refresh_from_db()
        self.assertTrue(self.rotation.is_closed)



import graphene
from graphene_django import DjangoObjectType

from allianceauth.fleetactivitytracking.models import Fatlink, Fat


class FatlinkType(DjangoObjectType):
    class Meta:
        model = Fatlink


class FatType(DjangoObjectType):
    class Meta:
        model = Fat


class FatUserStatsType(graphene.ObjectType):
    user = graphene.Field('allianceauth_graphql.authentication.types.UserType', required=True)
    num_chars = graphene.Int(required=True)
    num_fats = graphene.Int(required=True)
    average_fats = graphene.Float(required=True)


class FatCorpStatsType(graphene.ObjectType):
    corporation = graphene.Field('allianceauth_graphql.eveonline.types.EveCorporationInfoType', required=True)
    num_fats = graphene.Int(required=True)
    avg_fats = graphene.Float(required=True)


class FatPersonalStatsType(graphene.ObjectType):
    month = graphene.Int(required=True)
    year = graphene.Int(required=True)
    num_fats = graphene.Int(required=True)


class FatCollectedLinkType(graphene.ObjectType):
    shiptype = graphene.String(required=True)
    times_used = graphene.Int(required=True)


class FatPersonalMonthlyStatsType(graphene.ObjectType):
    collected_links = graphene.List(FatCollectedLinkType, required=True)
    created_links = graphene.List(FatlinkType, required=True)



import graphene
from graphene_django import DjangoObjectType

from allianceauth.timerboard.models import Timer, TimerType


TimerStructureChoices = graphene.Enum(
    'TimerStructureChoices',
    [
        ('POCO', 'POCO'),
        ('I_HUB', 'I-HUB'),
        ('TCU', 'TCU'),
        ('POS_S', 'POS[S]'),
        ('POS_M', 'POS[M]'),
        ('POS_L', 'POS[L]'),
        ('Astrahus', 'Astrahus'),
        ('Fortizar', 'Fortizar'),
        ('Keepstar', 'Keepstar'),
        ('Raitaru', 'Raitaru'),
        ('Azbel', 'Azbel'),
        ('Sotiyo', 'Sotiyo'),
        ('Athanor', 'Athanor'),
        ('Tatara', 'Tatara'),
        ('Pharolux_Cyno_Beacon', 'Pharolux Cyno Beacon'),
        ('Tenebrex_Cyno_Jammer', 'Tenebrex Cyno Jammer'),
        ('Ansiblex_Jump_Gate', 'Ansiblex Jump Gate'),
        ('Moon_Mining_Cycle', 'Moon Mining Cycle'),
        ('Other', 'Other')
    ]
)
TimerObjectiveChoices = graphene.Enum(
    'TimerObjectiveChoices',
    [
        ('Friendly', 'Friendly'),
        ('Hostile', 'Hostile'),
        ('Neutral', 'Neutral')
    ]
)
TimerTypeChoices = graphene.Enum.from_enum(TimerType)


class StructureTimerType(DjangoObjectType):
    timer_type = TimerTypeChoices()
    structure = TimerStructureChoices()
    objective = TimerObjectiveChoices()

    class Meta:
        model = Timer
        exclude = ('eve_corp', 'eve_character',)



import json
from graphene_django.utils.testing import GraphQLTestCase

from django.contrib.contenttypes.models import ContentType

from app_utils.testdata_factories import UserFactory
from allianceauth.tests.test_auth_utils import AuthUtils


class TestQueries(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserFactory()

        cls.perms = [
            ('permissions_tool', 'audit_permissions'),
            ('analytics', 'add_analyticsidentifier'),
            ('analytics', 'change_analyticsidentifier'),
            ('analytics', 'delete_analyticsidentifier'),
            ('analytics', 'view_analyticsidentifier'),
        ]

        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [f'{perm[0]}.{perm[1]}' for perm in cls.perms],
            cls.user,
            False
        )

    def test_show_only_applied_true(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q(
                $app_label: String
                $model: String
                $search_string: String
                $show_only_applied: Boolean
            ) {
                permsSearch(
                    appLabel: $app_label
                    model: $model
                    searchString: $search_string
                    showOnlyApplied: $show_only_applied
                ) {
                    codename
                    contentType {
                        appLabel
                    }
                }
            }
            ''',
            variables={'show_only_applied': True}
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('permsSearch', data['data'])

        results = data['data']['permsSearch']

        self.assertCountEqual([(r['contentType']['appLabel'], r['codename']) for r in results], self.perms)

    def test_app_label_not_only_applied(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q(
                $app_label: String
                $model: String
                $search_string: String
                $show_only_applied: Boolean
            ) {
                permsSearch(
                    appLabel: $app_label
                    model: $model
                    searchString: $search_string
                    showOnlyApplied: $show_only_applied
                ) {
                    codename
                    contentType {
                        appLabel
                    }
                }
            }
            ''',
            variables={
                'show_only_applied': False,
                'app_label': "permissions_tool",
            }
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('permsSearch', data['data'])

        results = data['data']['permsSearch']

        perms = [
            ('permissions_tool', 'add_permissionstool'),
            ('permissions_tool', 'change_permissionstool'),
            ('permissions_tool', 'delete_permissionstool'),
            ('permissions_tool', 'view_permissionstool'),
            ('permissions_tool', 'audit_permissions'),
        ]

        self.assertCountEqual([(r['contentType']['appLabel'], r['codename']) for r in results], perms)

    def test_model(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q(
                $app_label: String
                $model: String
                $search_string: String
                $show_only_applied: Boolean
            ) {
                permsSearch(
                    appLabel: $app_label
                    model: $model
                    searchString: $search_string
                    showOnlyApplied: $show_only_applied
                ) {
                    codename
                    contentType {
                        appLabel
                    }
                }
            }
            ''',
            variables={
                'show_only_applied': False,
                'model': "characterownership",
            }
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('permsSearch', data['data'])

        results = data['data']['permsSearch']

        perms = [
            ('authentication', 'change_characterownership'),
            ('authentication', 'delete_characterownership'),
        ]

        self.assertCountEqual([(r['contentType']['appLabel'], r['codename']) for r in results], perms)

    def test_search_string(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q(
                $app_label: String
                $model: String
                $search_string: String
                $show_only_applied: Boolean
            ) {
                permsSearch(
                    appLabel: $app_label
                    model: $model
                    searchString: $search_string
                    showOnlyApplied: $show_only_applied
                ) {
                    codename
                    contentType {
                        appLabel
                    }
                }
            }
            ''',
            variables={
                'show_only_applied': True,
                'search_string': "audit",
            }
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('permsSearch', data['data'])

        results = data['data']['permsSearch']

        perms = [
            ('permissions_tool', 'audit_permissions'),
        ]

        self.assertCountEqual([(r['contentType']['appLabel'], r['codename']) for r in results], perms)

    def test_perms_list_app_models(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q{
                permsListAppModels {
                    appLabel
                    model
                }
            }
            '''
        )

        data = json.loads(response.content)

        self.assertIn('data', data)
        self.assertNotIn('errors', data)
        self.assertIn('permsListAppModels', data['data'])

        results = data['data']['permsListAppModels']

        self.assertEqual(len(results), ContentType.objects.count())



import graphene


class TimerInput(graphene.InputObjectType):
    details = graphene.String(required=True)
    system = graphene.String(required=True)
    planet_moon = graphene.String(default_value="")
    structure = graphene.Field('allianceauth_graphql.timerboard.types.TimerStructureChoices', required=True)
    timer_type = graphene.Field('allianceauth_graphql.timerboard.types.TimerTypeChoices', required=True)
    objective = graphene.Field('allianceauth_graphql.timerboard.types.TimerObjectiveChoices', required=True)
    days_left = graphene.Int(required=True)
    hours_left = graphene.Int(required=True)
    minutes_left = graphene.Int(required=True)
    important = graphene.Boolean(default_value=False)
    corp_timer = graphene.Boolean(default_value=False)



from graphene_django.utils.testing import GraphQLTestCase

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory

from allianceauth.hrapplications.models import Application, ApplicationForm, ApplicationQuestion, ApplicationComment
from allianceauth.notifications.models import Notification

from ..hrapplications.types import ApplicationStatus


class TestQueriesAndTypes(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

        cls.character = cls.user.profile.main_character
        cls.corp = cls.character.corporation

        char2 = EveCharacterFactory(corporation=cls.corp)
        cls.user2 = UserMainFactory(main_character__character=char2)

        cls.corp2 = EveCorporationInfoFactory()

        cls.question1 = ApplicationQuestion.objects.create(title="Question 1")
        cls.choice1 = cls.question1.choices.create(choice_text="Choice 1")

        cls.form = ApplicationForm.objects.create(corp=cls.corp)
        cls.form.questions.add(cls.question1)

        cls.pending_application = Application.objects.create(
            user=cls.user,
            form=cls.form,
            approved=None
        )

        cls.form2 = ApplicationForm.objects.create(corp=cls.corp2)
        cls.form2.questions.add(cls.question1)

        cls.pending_application2 = Application.objects.create(
            user=cls.user,
            form=cls.form2,
            approved=None
        )

        cls.corp3 = EveCorporationInfoFactory()

        cls.form3 = ApplicationForm.objects.create(corp=cls.corp3)
        cls.form3.questions.add(cls.question1)

        cls.rejected_application = Application.objects.create(
            user=cls.user,
            form=cls.form3,
            approved=False
        )

        cls.approved_application = Application.objects.create(
            user=cls.user2,
            form=cls.form3,
            approved=True
        )

        cls.corp4 = EveCorporationInfoFactory()

        cls.form4 = ApplicationForm.objects.create(corp=cls.corp4)
        cls.form4.questions.add(cls.question1)

    def test_hr_corp_applications_superuser(self):
        self.client.force_login(self.user)

        self.user.is_superuser = True
        self.user.save()

        response = self.query(
            '''
            query {
                hrCorpApplications {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrCorpApplications": [
                        {
                            "id": str(self.pending_application2.id)
                        },
                        {
                            "id": str(self.pending_application.id)
                        },
                    ]
                }
            }
        )

    def test_hr_corp_applications_perms(self):
        user = AuthUtils.add_permission_to_user_by_name('auth.human_resources', self.user, False)

        self.client.force_login(user)

        response = self.query(
            '''
            query {
                hrCorpApplications {
                    id
                    status
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrCorpApplications": [
                        {
                            "id": str(self.pending_application.id),
                            "status": ApplicationStatus.PENDING.name
                        },
                    ]
                }
            }
        )

    def test_hr_corp_applications_no_access(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                hrCorpApplications {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrCorpApplications": []
                }
            }
        )

    def test_hr_finished_corp_applications_superuser(self):
        self.client.force_login(self.user)

        self.user.is_superuser = True
        self.user.save()

        response = self.query(
            '''
            query {
                hrFinishedCorpApplications {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrFinishedCorpApplications": [
                        {
                            "id": str(self.approved_application.id)
                        },
                        {
                            "id": str(self.rejected_application.id)
                        },
                    ]
                }
            }
        )

    def test_hr_finished_corp_applications_perms(self):
        user = AuthUtils.add_permission_to_user_by_name('auth.human_resources', self.user, False)

        self.client.force_login(user)

        self.character.corporation_id = self.corp3.corporation_id
        self.character.save()

        response = self.query(
            '''
            query {
                hrFinishedCorpApplications {
                    id
                    status
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrFinishedCorpApplications": [
                        {
                            "id": str(self.approved_application.id),
                            "status": ApplicationStatus.APPROVED.name
                        },
                        {
                            "id": str(self.rejected_application.id),
                            "status": ApplicationStatus.REJECTED.name
                        },
                    ]
                }
            }
        )

    def test_hr_finished_corp_applications_no_access(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                hrFinishedCorpApplications {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrFinishedCorpApplications": []
                }
            }
        )

    def test_hr_list_available_forms(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                hrListAvailableForms {
                    id
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrListAvailableForms": [
                        {
                            "id": str(self.form4.id),
                        },
                    ]
                }
            }
        )

    def test_hr_personal_applications_all(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query {
                hrPersonalApplications {
                    id
                    status
                }
            }
            '''
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrPersonalApplications": [
                        {
                            "id": str(self.pending_application.id),
                            "status": ApplicationStatus.PENDING.name
                        },
                        {
                            "id": str(self.pending_application2.id),
                            "status": ApplicationStatus.PENDING.name
                        },
                        {
                            "id": str(self.rejected_application.id),
                            "status": ApplicationStatus.REJECTED.name
                        },
                    ]
                }
            }
        )

    def test_hr_personal_applications_pending(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($status: ApplicationStatus) {
                hrPersonalApplications(status: $status) {
                    id
                }
            }
            ''',
            variables={
                "status": ApplicationStatus.PENDING.name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrPersonalApplications": [
                        {
                            "id": str(self.pending_application.id)
                        },
                        {
                            "id": str(self.pending_application2.id)
                        },
                    ]
                }
            }
        )

    def test_hr_personal_applications_rejected(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            query($status: ApplicationStatus) {
                hrPersonalApplications(status: $status) {
                    id
                }
            }
            ''',
            variables={
                "status": ApplicationStatus.REJECTED.name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrPersonalApplications": [
                        {
                            "id": str(self.rejected_application.id)
                        },
                    ]
                }
            }
        )

    def test_hr_personal_applications_approved(self):
        self.client.force_login(self.user2)

        response = self.query(
            '''
            query($status: ApplicationStatus) {
                hrPersonalApplications(status: $status) {
                    id
                    status
                }
            }
            ''',
            variables={
                "status": ApplicationStatus.APPROVED.name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrPersonalApplications": [
                        {
                            "id": str(self.approved_application.id),
                            "status": ApplicationStatus.APPROVED.name
                        },
                    ]
                }
            }
        )

    def test_hr_search_application_superuser(self):
        self.client.force_login(self.user)

        self.user.is_superuser = True
        self.user.save()

        response = self.query(
            '''
            query($searchString: String!) {
                hrSearchApplication(searchString: $searchString) {
                    id
                    status
                }
            }
            ''',
            variables={
                "searchString": self.corp.corporation_name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrSearchApplication": [
                        {
                            'id': str(self.pending_application.id),
                            'status': ApplicationStatus.PENDING.name
                        },
                        {
                            'id': str(self.pending_application2.id),
                            'status': ApplicationStatus.PENDING.name
                        },
                        {
                            "id": str(self.rejected_application.id),
                            "status": ApplicationStatus.REJECTED.name
                        },
                        {
                            'id': str(self.approved_application.id),
                            'status': ApplicationStatus.APPROVED.name
                        }
                    ]
                }
            }
        )

    def test_hr_search_application_perms(self):
        user = AuthUtils.add_permission_to_user_by_name('auth.human_resources', self.user, False)

        self.client.force_login(user)

        response = self.query(
            '''
            query($searchString: String!) {
                hrSearchApplication(searchString: $searchString) {
                    id
                }
            }
            ''',
            variables={
                "searchString": self.corp3.corporation_name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrSearchApplication": []
                }
            }
        )

    def test_hr_search_application_error(self):
        user = AuthUtils.add_permission_to_user_by_name('auth.human_resources', self.user, False)

        self.client.force_login(user)

        self.character.delete()

        response = self.query(
            '''
            query($searchString: String!) {
                hrSearchApplication(searchString: $searchString) {
                    id
                }
            }
            ''',
            variables={
                "searchString": self.corp3.corporation_name
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrSearchApplication": None
                }
            }
        )


class TestCreateApplicationMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

        cls.corp = cls.user.profile.main_character.corporation

        cls.question1 = ApplicationQuestion.objects.create(title="Question 1")
        cls.choice1 = cls.question1.choices.create(choice_text="Choice 1")

        cls.form = ApplicationForm.objects.create(corp=cls.corp)
        cls.form.questions.add(cls.question1)

        cls.form_data = {
            'formId': cls.form.pk,
            'responses': [
                {
                    'questionId': cls.question1.pk,
                    'answer': ['Choice 1']
                }
            ]
        }

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($input: FormAnswerInputType!) {
                hrCreateApplication(input: $input) {
                    ok
                    application {
                        form {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                "data": {
                    "hrCreateApplication": {
                        "ok": True,
                        "application": {
                            'form': {
                                'id': str(self.form.pk)
                            }
                        }
                    }
                }
            }
        )

    def test_already_existing(self):
        self.client.force_login(self.user)

        Application.objects.create(
            user=self.user,
            form=self.form
        )

        response = self.query(
            '''
            mutation($input: FormAnswerInputType!) {
                hrCreateApplication(input: $input) {
                    ok
                    application {
                        form {
                            id
                        }
                    }
                }
            }
            ''',
            input_data=self.form_data
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrCreateApplication': {
                        'ok': False,
                        'application': None
                    }
                }
            }
        )


class TestDeleteApplicationMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

        corp = cls.user.profile.main_character.corporation

        form = ApplicationForm.objects.create(corp=corp)

        cls.application: Application = Application.objects.create(
            form=form,
            user=cls.user,
            approved=True
        )

    def test_ok(self):
        self.client.force_login(self.user)

        self.application.approved = None
        self.application.save()

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrDeleteApplication(applicationId: $applicationId) {
                    ok
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrDeleteApplication': {
                        'ok': True
                    }
                }
            }
        )

        self.assertEqual(Application.objects.count(), 0)

    def test_not_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrDeleteApplication(applicationId: $applicationId) {
                    ok
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrDeleteApplication': {
                        'ok': False
                    }
                }
            }
        )

        self.assertEqual(Application.objects.count(), 1)


class TestAddApplicationCommentMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'auth.human_resources',
                'hrapplications.add_applicationcomment'
            ],
            UserMainFactory(),
            False
        )

        corp = cls.user.profile.main_character.corporation

        form = ApplicationForm.objects.create(corp=corp)

        cls.application: Application = Application.objects.create(
            form=form,
            user=cls.user,
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!, $comment: String!) {
                hrAddApplicationComment(applicationId: $applicationId, comment: $comment) {
                    ok
                    application {
                        comments {
                            text
                        }
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
                'comment': "Test comment"
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAddApplicationComment': {
                        'ok': True,
                        'application': {
                            'comments': [
                                {
                                    'text': "Test comment"
                                }
                            ]
                        }
                    }
                }
            }
        )

        self.assertEqual(ApplicationComment.objects.count(), 1)


class TestAdminRemoveApplicationMutation(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'auth.human_resources',
                'hrapplications.delete_application'
            ],
            UserMainFactory(),
            False
        )

        corp = cls.user.profile.main_character.corporation

        form = ApplicationForm.objects.create(corp=corp)

        cls.application: Application = Application.objects.create(
            form=form,
            user=cls.user,
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminRemoveApplication(applicationId: $applicationId) {
                    ok
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminRemoveApplication': {
                        'ok': True,
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 1)

        self.assertEqual(Application.objects.count(), 0)


class TestAdminApproveAndRejectApplication(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permissions_to_user_by_name(
            [
                'auth.human_resources',
                'hrapplications.approve_application',
                'hrapplications.reject_application'
            ],
            UserMainFactory(),
            False
        )

        cls.user2 = UserMainFactory()

        corp = cls.user2.profile.main_character.corporation

        form = ApplicationForm.objects.create(corp=corp)

        cls.application: Application = Application.objects.create(
            form=form,
            user=cls.user2
        )

    def test_approve_ok(self):
        self.client.force_login(self.user)

        self.user.is_superuser = True
        self.user.save()

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminApproveApplication(applicationId: $applicationId) {
                    ok
                    application {
                        status
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminApproveApplication': {
                        'ok': True,
                        'application': {
                            'status': ApplicationStatus.APPROVED.name
                        }
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 1)

    def test_approve_not_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminApproveApplication(applicationId: $applicationId) {
                    ok
                    application {
                        status
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminApproveApplication': {
                        'ok': False,
                        'application': {
                            'status': ApplicationStatus.PENDING.name
                        }
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 0)

    def test_reject_ok(self):
        self.client.force_login(self.user)

        self.user.is_superuser = True
        self.user.save()

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminRejectApplication(applicationId: $applicationId) {
                    ok
                    application {
                        status
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminRejectApplication': {
                        'ok': True,
                        'application': {
                            'status': ApplicationStatus.REJECTED.name
                        }
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 1)

    def test_reject_not_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminRejectApplication(applicationId: $applicationId) {
                    ok
                    application {
                        status
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminRejectApplication': {
                        'ok': False,
                        'application': {
                            'status': ApplicationStatus.PENDING.name
                        }
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 0)


class TestAdminMarkInProgressApplication(GraphQLTestCase):
    maxDiff = None

    @classmethod
    def setUpTestData(cls):
        cls.user = AuthUtils.add_permission_to_user_by_name('auth.human_resources', UserMainFactory(), False)

        cls.user2 = UserMainFactory()

        corp = cls.user2.profile.main_character.corporation

        form = ApplicationForm.objects.create(corp=corp)

        cls.application: Application = Application.objects.create(
            form=form,
            user=cls.user2
        )

    def test_ok(self):
        self.client.force_login(self.user)

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminMarkInProgressApplication(applicationId: $applicationId) {
                    ok
                    application {
                        reviewer {
                            id
                        }
                        reviewerCharacter {
                            id
                        }
                    }
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminMarkInProgressApplication': {
                        'ok': True,
                        'application': {
                            'reviewer': {
                                'id': str(self.user.id)
                            },
                            'reviewerCharacter': {
                                'id': str(self.user.profile.main_character.id)
                            }
                        }
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 1)

    def test_not_ok(self):
        self.client.force_login(self.user)

        self.application.reviewer = self.user
        self.application.save()

        response = self.query(
            '''
            mutation($applicationId: ID!) {
                hrAdminMarkInProgressApplication(applicationId: $applicationId) {
                    ok
                }
            }
            ''',
            variables={
                'applicationId': self.application.pk,
            }
        )

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'hrAdminMarkInProgressApplication': {
                        'ok': False,
                    }
                }
            }
        )

        self.assertEqual(Notification.objects.count(), 0)



import graphene
from graphql_jwt.decorators import login_required, permission_required

from django.utils import timezone

from allianceauth.optimer.models import OpTimer

from .types import OpTimerModelType


class Query:
    optimer_past_timers = graphene.List(OpTimerModelType)
    optimer_future_timers = graphene.List(OpTimerModelType)

    @login_required
    @permission_required('auth.optimer_view')
    def resolve_optimer_past_timers(self, info):
        return OpTimer.objects.filter(start__lt=timezone.now()).order_by('-start')

    @login_required
    @permission_required('auth.optimer_view')
    def resolve_optimer_future_timers(self, info):
        return OpTimer.objects.filter(start__gte=timezone.now())



import graphene
from graphql_jwt.decorators import login_required, user_passes_test

from django.db.models import Case, When, Exists, OuterRef, Q, Count
from django.conf import settings
from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied

from allianceauth.groupmanagement.managers import GroupManager
from allianceauth.groupmanagement.models import GroupRequest, RequestLog
from allianceauth.services.hooks import get_extension_logger

from allianceauth_graphql.authentication.types import GroupType

from .types import GroupManagementType, GroupMembershipListType, GroupMembershipAuditType


logger = get_extension_logger(__name__)


class Query:
    groupmanagement_user_joinable_groups = graphene.List(GroupType)
    groupmanagement_manage_requests = graphene.Field(GroupManagementType)
    groupmanagement_groups = graphene.List(GroupType)
    groupmanagement_group_memberships = graphene.Field(GroupMembershipListType, group_id=graphene.Int(required=True))
    groupmanagement_group_membership_audit = graphene.Field(GroupMembershipAuditType, group_id=graphene.Int(required=True))

    @login_required
    def resolve_groupmanagement_user_joinable_groups(self, info):
        user = info.context.user
        groups_qs = GroupManager.get_joinable_groups_for_user(user, include_hidden=False).order_by('name')

        return groups_qs.annotate(
            status=Case(
                When(
                    Exists(user.groups.filter(pk=OuterRef('pk'))),
                    then=Case(
                        When(
                            Exists(GroupRequest.objects.filter(user=user, group_id=OuterRef('pk'))),
                            then=2
                        ),
                        default=1
                    )
                ),
                When(
                    ~Exists(GroupRequest.objects.filter(user=user, group_id=OuterRef('pk'))),
                    then=Case(
                        When(
                            Q(authgroup__open=True), then=3
                        ),
                        default=4
                    )
                ),
                default=2
            )
        )

    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def resolve_groupmanagement_manage_requests(self, info):
        user = info.context.user
        logger.debug(f"group_management called by user {user}")
        acceptrequests = []
        leaverequests = []

        if GroupManager.has_management_permission(user):
            # Full access
            group_requests = GroupRequest.objects.all()
        else:
            # Group specific leader
            users__groups = GroupManager.get_group_leaders_groups(user)
            group_requests = GroupRequest.objects.filter(group__in=users__groups)

        for grouprequest in group_requests:
            if grouprequest.leave_request:
                leaverequests.append(grouprequest)
            else:
                acceptrequests.append(grouprequest)

        logger.debug(f"Providing user {user} with {len(acceptrequests)} acceptrequests and {len(leaverequests)} leaverequests.")

        return {
            'leave_requests': leaverequests,
            'accept_requests': acceptrequests,
            'auto_leave': getattr(settings, 'GROUPMANAGEMENT_AUTO_LEAVE', False),
        }

    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def resolve_groupmanagement_groups(self, info):
        user = info.context.user
        logger.debug(f"group_membership called by user {user}")
        # Get all open and closed groups
        if GroupManager.has_management_permission(user):
            # Full access
            groups = GroupManager.get_all_non_internal_groups()
        else:
            # Group leader specific
            groups = GroupManager.get_group_leaders_groups(user)

        return (
            groups
            .exclude(authgroup__internal=True)
            .annotate(num_members=Count('user'))
            .order_by('name')
        )

    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def resolve_groupmanagement_group_memberships(self, info, group_id):
        user = info.context.user
        logger.debug(f"group_membership_list called by user {user} for group id {group_id}")
        try:
            group = Group.objects.get(id=group_id)
            # Check its a joinable group i.e. not corp or internal
            # And the user has permission to manage it
            if (not GroupManager.check_internal_group(group)
                or not GroupManager.can_manage_group(user, group)
                ):
                logger.warning(
                    f"User {user} attempted to view the membership of group {group_id} "
                    "but permission was denied"
                )
                raise PermissionDenied

        except ObjectDoesNotExist:
            raise Exception("Group doesn't exist")

        group_leaders = group.authgroup.group_leaders.all()
        members = list()
        for member in \
            group.user_set\
                .all()\
                .order_by('profile__main_character__character_name'):

            members.append({
                'user': member,
                'is_leader': member in group_leaders
            })

        return {'group': group, 'members': members}

    @login_required
    @user_passes_test(GroupManager.can_manage_groups)
    def resolve_groupmanagement_group_membership_audit(self, info, group_id):
        user = info.context.user
        logger.debug("group_management_audit called by user %s" % user)
        try:
            group = Group.objects.get(id=group_id)
            # Check its a joinable group i.e. not corp or internal
            # And the user has permission to manage it
            if not GroupManager.check_internal_group(group) or not GroupManager.can_manage_group(user, group):
                logger.warning(f"User {user} attempted to view the membership of group {group_id} but permission was denied")
                raise PermissionDenied

        except ObjectDoesNotExist:
            raise Exception("Group does not exist")

        entries = RequestLog.objects.filter(group=group).order_by('-date')

        return {'group': group, 'entries': entries}



import graphene
from graphene_django.forms.mutation import DjangoFormMutation
from graphql_jwt.decorators import login_required, permission_required

from allianceauth.optimer.form import OpForm
from allianceauth.optimer.models import OpTimer, OpTimerType

from .forms import EditOpForm
from .types import OpTimerModelType


class OpFormMutation(DjangoFormMutation):
    class Meta:
        form_class = OpForm

    op = graphene.Field(OpTimerModelType)

    @classmethod
    @login_required
    @permission_required('auth.optimer_management')
    def perform_mutate(cls, form: OpForm, info):
        optimer_type = None
        if form.cleaned_data['type'] != '':
            try:
                optimer_type = OpTimerType.objects.get(
                    type__iexact=form.cleaned_data['type']
                )
            except OpTimerType.DoesNotExist:
                optimer_type = OpTimerType.objects.create(
                    type=form.cleaned_data['type']
                )

        character = info.context.user.profile.main_character
        op = OpTimer()
        op.doctrine = form.cleaned_data['doctrine']
        op.system = form.cleaned_data['system']
        op.start = form.cleaned_data['start']
        op.duration = form.cleaned_data['duration']
        op.operation_name = form.cleaned_data['operation_name']
        op.fc = form.cleaned_data['fc']
        op.eve_character = character
        op.type = optimer_type
        op.description = form.cleaned_data['description']
        op.save()

        return cls(op=op, **form.cleaned_data)


class RemoveOpTimerMutation(graphene.Mutation):
    class Arguments:
        optimer_id = graphene.ID(required=True)

    ok = graphene.Boolean()

    @classmethod
    @login_required
    @permission_required('auth.optimer_management')
    def mutate(cls, root, info, optimer_id):
        op = OpTimer.objects.get(pk=optimer_id)
        op.delete()
        return cls(ok=True)


class OpFormEditMutation(DjangoFormMutation):
    class Meta:
        form_class = EditOpForm

    op = graphene.Field(OpTimerModelType)

    @classmethod
    @login_required
    @permission_required('auth.optimer_management')
    def perform_mutate(cls, form: EditOpForm, info):
        op: OpTimer = OpTimer.objects.get(pk=form.cleaned_data['op_id'])

        character = info.context.user.profile.main_character

        optimer_type = None

        if form.cleaned_data['type'] != '':
            try:
                optimer_type = OpTimerType.objects.get(
                    type__iexact=form.cleaned_data['type']
                )
            except OpTimerType.DoesNotExist:
                optimer_type = OpTimerType.objects.create(
                    type=form.cleaned_data['type']
                )

        op.doctrine = form.cleaned_data['doctrine']
        op.system = form.cleaned_data['system']
        op.start = form.cleaned_data['start']
        op.duration = form.cleaned_data['duration']
        op.operation_name = form.cleaned_data['operation_name']
        op.fc = form.cleaned_data['fc']
        op.eve_character = character
        op.type = optimer_type
        op.description = form.cleaned_data['description']
        op.save()

        return cls(op=op, **form.cleaned_data)


class Mutation:
    optimer_new_op = OpFormMutation.Field()
    optimer_remove_op = RemoveOpTimerMutation.Field()
    optimer_edit_op = OpFormEditMutation.Field()



from graphene_django.utils.testing import GraphQLTestCase

from app_utils.testdata_factories import UserMainFactory


class TestAll(GraphQLTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_user_tokens(self):
        self.client.force_login(self.user, "graphql_jwt.backends.JSONWebTokenBackend")

        response = self.query(
            '''
            query q {
                esiUserTokens {
                    id
                    character {
                        id
                    }
                }
            }
            '''
        )

        self.assertEqual(self.user.token_set.count(), 1)

        token = self.user.token_set.first()

        self.assertJSONEqual(
            response.content,
            {
                'data': {
                    'esiUserTokens': [
                        {
                            'id': str(token.pk),
                            'character': {
                                'id': str(self.user.profile.main_character.pk)
                            }
                        }
                    ]
                }
            }
        )



from django.contrib.auth.models import Permission, Group
from django.contrib.contenttypes.models import ContentType

import graphene
from graphene_django import DjangoObjectType

from allianceauth.authentication.models import State


class ContentTypeType(DjangoObjectType):
    class Meta:
        model = ContentType
        fields = ('app_label', 'model',)


class GroupAdminType(DjangoObjectType):
    class Meta:
        model = Group
        fields = ('name', 'authgroup', 'id', 'user_set',)


class StateAdminType(DjangoObjectType):
    class Meta:
        model = State
        fields = ('name', 'userprofile_set',)


class PermissionType(DjangoObjectType):
    num_users = graphene.Int()
    num_groups = graphene.Int()
    num_users_in_groups = graphene.Int()
    num_states = graphene.Int()
    num_users_in_states = graphene.Int()

    group_set = graphene.List(GroupAdminType, required=True)
    state_set = graphene.List(StateAdminType, required=True)

    class Meta:
        model = Permission
        fields = (
            'id',
            'name',
            'codename',
            'content_type',
            'user_set',
            'state_set',
            'group_set',
        )


class AppModelType(graphene.ObjectType):
    app_label = graphene.String(required=True)
    model = graphene.String(required=True)
